import os

import pytest
from fastapi.testclient import TestClient

from recce.core import default_context
from recce.server import app
# noinspection PyUnresolvedReferences
from tests.adapter.dbt_adapter.conftest import dbt_test_helper


@pytest.fixture
def temp_folder():
    import tempfile
    temp_dir = tempfile.mkdtemp()
    yield temp_dir

    import shutil
    shutil.rmtree(temp_dir)


def test_health():
    client = TestClient(app)
    response = client.get("/api/health")
    assert response.status_code == 200
    assert response.json() == {"status": "ok"}


def test_stateless(dbt_test_helper):
    context = default_context()
    from recce.state import RecceStateLoader
    context.state_loader = RecceStateLoader()
    client = TestClient(app)
    response = client.get("/api/info")
    assert response.status_code == 200
    info = response.json()
    assert info['file_mode'] is False
    assert info['cloud_mode'] is False


def test_file_mode(dbt_test_helper):
    context = default_context()
    from recce.state import RecceStateLoader
    context.state_loader = RecceStateLoader(state_file='/tmp/recce_state.json')
    client = TestClient(app)
    response = client.get("/api/info")
    assert response.status_code == 200
    info = response.json()
    assert info['file_mode'] is True
    assert info['filename'] == 'recce_state.json'
    assert info['cloud_mode'] is False


def test_saveas_and_rename(dbt_test_helper, temp_folder):
    context = default_context()
    state_file = os.path.join(temp_folder, 'recce_state.json')
    state_file2 = os.path.join(temp_folder, 'recce_state2.json')
    state_file3 = os.path.join(temp_folder, 'recce_state3.json')
    os.makedirs(os.path.join(temp_folder, 'dir.json'))

    from recce.state import RecceStateLoader
    context.state_loader = RecceStateLoader(state_file=state_file)
    client = TestClient(app)

    response = client.post("/api/save", json={"filename": "recce_state2.json"})
    assert response.status_code == 200
    assert os.path.exists(state_file)

    response = client.post("/api/save-as", json={"filename": "recce_state2.json"})
    assert response.status_code == 200
    assert os.path.exists(state_file2)
    assert context.state_loader.state_file == os.path.join(temp_folder, 'recce_state2.json')

    # Same file
    response = client.post("/api/save-as", json={"filename": "recce_state2.json"})
    assert response.status_code == 400

    # folder
    response = client.post("/api/save-as", json={"filename": "dir.json"})
    assert response.status_code == 400

    # Rename
    response = client.post("/api/rename", json={"filename": "recce_state3.json"})
    assert response.status_code == 200
    assert not os.path.exists(state_file2)
    assert os.path.exists(state_file3)
    assert context.state_loader.state_file == os.path.join(temp_folder, 'recce_state3.json')

    # Conflict
    response = client.post("/api/save-as", json={"filename": "recce_state.json"})
    assert response.status_code == 409
    response = client.post("/api/rename", json={"filename": "recce_state.json"})
    assert response.status_code == 409

    # Overwrite
    response = client.post("/api/save-as", json={"filename": "recce_state.json", "overwrite": True})
    assert response.status_code == 200
    assert context.state_loader.state_file == os.path.join(temp_folder, 'recce_state.json')



from recce.adapter.dbt_adapter import DbtAdapter, dbt_supported_registry
from recce.models import RunType


def test_dbt_adapter_support_tasks(dbt_test_helper):
    adapter: DbtAdapter = dbt_test_helper.context.adapter

    # Test dbt package name
    supported_dbt_packages = set([package.package_name for package in adapter.manifest.macros.values()])
    assert 'audit_helper' in supported_dbt_packages
    assert 'dbt_profiler' in supported_dbt_packages

    # Test dbt task support
    support_tasks = adapter.support_tasks()

    for task_type in dbt_supported_registry:
        task = task_type.value
        assert task in support_tasks
        assert support_tasks[task] is True


def test_dbt_adapter_support_tasks_without_required_dbt_package(dbt_test_helper):
    adapter: DbtAdapter = dbt_test_helper.context.adapter
    # Mock the macros in manifest to simulate no required dbt package installed
    adapter.manifest.macros = {}

    support_tasks = adapter.support_tasks()

    for task_type in dbt_supported_registry:
        task = task_type.value
        assert task in support_tasks
        if task == RunType.PROFILE_DIFF.value:
            assert support_tasks[task] is False
        elif task == RunType.VALUE_DIFF.value:
            assert support_tasks[task] is False
        elif task == RunType.VALUE_DIFF_DETAIL.value:
            assert support_tasks[task] is False

    # Check the query_diff_with_primary_key is not supported
    assert support_tasks["query_diff_with_primary_key"] is False



# Configuration file for recce
from rich.console import Console

from recce import yaml
from recce.exceptions import RecceConfigException
from recce.util import SingletonMeta

RECCE_CONFIG_FILE = 'recce.yml'
RECCE_PRESET_CHECK_COMMENT = '''Preset Checks
Please see https://datarecce.io/docs/features/preset-checks/
'''
RECCE_ERROR_LOG_FILE = 'recce_error.log'
console = Console()


class RecceConfig(metaclass=SingletonMeta):
    def __init__(self, config_file=RECCE_CONFIG_FILE):
        self.config_file = config_file
        self.config = None
        self.load()

    def load(self):
        try:
            with open(self.config_file, 'r') as f:
                self.config = yaml.safe_load(f)
            self._verify_preset_checks()
        except FileNotFoundError:
            console.print(f'Recce config file not found. Generating default config file at \'{self.config_file}\'')
            self.config = self.generate_template()
            self.save()

    def _verify_preset_checks(self):
        from recce.tasks.core import CheckValidator

        if 'checks' not in self.config:
            return

        for check in self.config['checks']:
            try:
                check_type = check.get('type')
                if check_type is None:
                    raise ValueError(f'Check type is required for check "{check}"')
                if check_type == 'lineage_diff':
                    from recce.tasks.lineage import LineageDiffCheckValidator
                    validator = LineageDiffCheckValidator()
                elif check_type == 'schema_diff':
                    from recce.tasks.schema import SchemaDiffCheckValidator
                    validator = SchemaDiffCheckValidator()
                elif check_type == 'row_count_diff':
                    from recce.tasks.rowcount import RowCountDiffCheckValidator
                    validator = RowCountDiffCheckValidator()
                elif check_type == 'query':
                    from recce.tasks.query import QueryCheckValidator
                    validator = QueryCheckValidator()
                elif check_type == 'query_diff':
                    from recce.tasks.query import QueryDiffCheckValidator
                    validator = QueryDiffCheckValidator()
                elif check_type == 'value_diff' or check_type == 'value_diff_detail':
                    from recce.tasks.valuediff import ValueDiffCheckValidator
                    validator = ValueDiffCheckValidator()
                elif check_type == 'profile_diff':
                    from recce.tasks.profile import ProfileDiffCheckValidator
                    validator = ProfileDiffCheckValidator()
                elif check_type == 'top_k_diff':
                    from recce.tasks.top_k import TopKDiffCheckValidator
                    validator = TopKDiffCheckValidator()
                elif check_type == 'histogram_diff':
                    from recce.tasks.histogram import HistogramDiffCheckValidator
                    validator = HistogramDiffCheckValidator()
                else:
                    validator = CheckValidator()
                validator.validate(check)
            except Exception as e:
                import json
                raise RecceConfigException(
                    f"Load preset checks failed from '{self.config_file}'\n{json.dumps(check, indent=2)}",
                    cause=e)

    def generate_template(self):
        data = yaml.CommentedMap(
            checks=yaml.CommentedSeq())
        data.yaml_set_comment_before_after_key('checks', before=RECCE_PRESET_CHECK_COMMENT)
        # Define default preset checks
        default_checks = [
            yaml.CommentedMap(
                name='Row count diff',
                description='Check the row count diff for all table models.',
                type='row_count_diff',
                params={'select': 'state:modified,config.materialized:table'},
            ),
            yaml.CommentedMap(
                name='Schema diff',
                description='Check the schema diff for all nodes.',
                type='schema_diff',
            )
        ]

        for check in default_checks:
            data['checks'].append(check)

        return data

    def get(self, key, default=None):
        return self.config.get(key, default)

    def set(self, key, value):
        self.config[key] = value

    def save(self):
        with open(RECCE_CONFIG_FILE, 'w') as f:
            yaml.dump(self.config, f)

    def __str__(self):
        return yaml.dump(self.config)

    def __repr__(self):
        return self.__str__()



from typing import Any, Callable

from ruamel import yaml
from ruamel.yaml import CommentedMap as _cm, CommentedSeq as _cs

_yaml = yaml.YAML()
_safe_yaml = yaml.YAML(typ='safe')

CommentedMap = _cm
CommentedSeq = _cs
YAMLError = yaml.YAMLError


def load(stream) -> Any:
    return _yaml.load(stream)


def allow_duplicate_keys_loader() -> Callable:
    yml = yaml.YAML()
    yml.allow_duplicate_keys = True
    return yml.load


def safe_load(stream, version=None) -> Any:
    if version is not None:
        _safe_yaml.version = version
    return _safe_yaml.load(stream)


def dump(
        data, stream: Any = None, *, transform: Any = None
) -> Any:
    return _yaml.dump(data, stream, transform=transform)


def safe_load_yaml(file_path):
    try:
        with open(file_path, 'r') as f:
            payload = safe_load(f)
    except yaml.YAMLError as e:
        print(e)
        return None
    except FileNotFoundError:
        return None
    return payload


def round_trip_load_yaml(file_path):
    with open(file_path, 'r') as f:
        try:
            payload = load(f)
        except yaml.YAMLError as e:
            print(e)
            return None
    return payload


def round_trip_dump(
        data: Any,
        stream=None):
    return yaml.round_trip_dump(data, stream)



import os


def is_ci_env():
    # List of CI environment variables and their expected values
    ci_environments = {
        'CI': 'true',                   # Generic CI indicator
        'CIRCLECI': 'true',             # CircleCI
        'GITHUB_ACTIONS': 'true',       # GitHub Actions
        'GITLAB_CI': 'true',            # GitLab CI
        'JENKINS_URL': None,            # Jenkins (just needs to exist)
        'TRAVIS': 'true',               # Travis CI
        'APPVEYOR': 'true',             # AppVeyor
        'DRONE': 'true',                # Drone CI
        'TEAMCITY_VERSION': None,       # TeamCity
        'BITBUCKET_COMMIT': None,       # Bitbucket Pipelines
        'BUILDKITE': 'true',            # Buildkite
        'CODEBUILD_BUILD_ID': None,     # AWS CodeBuild
        'AZURE_PIPELINES': 'true',      # Azure Pipelines
    }

    for env_var, expected_value in ci_environments.items():
        env_value = os.environ.get(env_var)
        if env_value is not None:
            # If we just need the variable to exist
            if expected_value is None:
                return True
            # If we need to match a specific value (case-insensitive)
            if env_value.lower() == expected_value.lower():
                return True

    return False


def get_runner():
    # GitHub Action
    if os.environ.get('GITHUB_ACTIONS', 'false') == 'true':
        return 'github actions'

    # GitHub Codespace
    if os.environ.get('CODESPACES', 'false') == 'true':
        return 'github codespaces'

    # CircleCI
    if os.environ.get('CIRCLECI', 'false') == 'true':
        return 'circleci'

    return None


def get_version():
    version_file = os.path.normpath(os.path.join(os.path.dirname(__file__), 'VERSION'))
    with open(version_file) as fh:
        version = fh.read().strip()
        return version


__version__ = get_version()



from .types import Run, RunType


class RunDAO:
    """
    Data Access Object for Run. Currently, we store runs in memory, in the future, we can store them in a database.
    """

    @property
    def _runs(self):
        from recce.core import default_context
        return default_context().runs

    def create(self, run: Run):
        self._runs.append(run)

    def find_run_by_id(self, run_id):
        for run in self._runs:
            if str(run_id) == str(run.run_id):
                return run

        return None

    def list(self, type_filter: RunType = None):
        if type_filter:
            return list(filter(lambda run: run.type == type_filter, self._runs))
        return list(self._runs)

    def list_by_check_id(self, check_id):
        runs = []
        for run in self._runs:
            if str(check_id) == str(run.check_id):
                runs.append(run)
        return runs

    def delete(self, run_id):
        for run in self._runs:
            if str(run_id) == str(run.run_id):
                self._runs.remove(run)
                return True

        return False

    def clear(self):
        self._runs.clear()



import pytest

from recce.core import set_default_context
from .dbt_test_helper import DbtTestHelper


@pytest.fixture
def dbt_test_helper():
    helper = DbtTestHelper()
    context = helper.context
    set_default_context(context)
    yield helper
    helper.cleanup()



from typing import Union, List, Optional, Literal

from pydantic import BaseModel

from recce.models import Check
from recce.tasks.core import TaskResultDiffer, CheckValidator


class SchemaDiffResultDiffer:
    related_node_ids: List[str] = None

    def __init__(self, check, base_lineage, curr_lineage):
        self.check = check
        self.related_node_ids = self._get_related_node_ids()
        self.changes = self._check_result_changed_fn(base_lineage, curr_lineage)
        self.changed_nodes = self._get_changed_nodes()

    def _get_related_node_ids(self) -> Union[List[str], None]:
        params = self.check.params
        if params.get('node_id'):
            return params.get('node_id') if isinstance(params.get('node_id'), list) else [params.get('node_id')]
        else:
            return TaskResultDiffer.get_node_ids_by_selector(
                select=params.get('select'),
                exclude=params.get('exclude'),
                packages=params.get('packages'),
                view_mode=params.get('view_mode'),
            )

    def _check_result_changed_fn(self, base_lineage, curr_lineage):
        base = {}
        current = {}
        base_nodes = base_lineage.get('nodes', {})
        curr_nodes = curr_lineage.get('nodes', {})
        for node_id in self.related_node_ids:
            node = curr_nodes.get(node_id) or base_nodes.get(node_id)
            if not node:
                continue

            node_name = node.get('name')
            base[node_name] = base_nodes.get(node_id, {}).get('columns', {})
            current[node_name] = curr_nodes.get(node_id, {}).get('columns', {})

        return TaskResultDiffer.diff(base, current)

    def _get_changed_nodes(self) -> Union[List[str], None]:
        if self.changes:
            return self.changes.affected_root_keys.items


class SchemaDiffParams(BaseModel):
    node_id: Optional[Union[str, List[str]]] = None
    select: Optional[str] = None
    exclude: Optional[str] = None
    packages: Optional[list[str]] = None
    view_mode: Optional[Literal['all', 'changed_models']] = None


class SchemaDiffCheckValidator(CheckValidator):
    def validate_check(self, check: Check):
        SchemaDiffParams(**check.params)
        SchemaDiffParams(**check.view_options)



import os
import sys
import time
from datetime import datetime, timezone
from typing import List

from deepdiff import DeepDiff
from rich import box
from rich.console import Console
from rich.table import Table

from recce.apis.check_func import create_check_from_run, create_check_without_run, purge_preset_checks
from recce.apis.run_func import submit_run
from recce.config import RecceConfig
from recce.core import default_context
from recce.models.types import RunType
from recce.summary import generate_markdown_summary


def check_github_ci_env(**kwargs):
    """Check if the environment is GitHub CI"""
    if 'GITHUB_ACTIONS' not in os.environ:
        return False, None

    # Get the PR number
    github_server_url = os.environ.get('GITHUB_SERVER_URL', 'https://github.com')
    github_repository = os.environ.get('GITHUB_REPOSITORY', '')
    github_ref_name = os.environ.get('GITHUB_REF_NAME', '')
    github_pull_request_url = f"{github_server_url}/{github_repository}/pull/{github_ref_name}"

    return True, github_pull_request_url


def load_preset_checks(checks: list):
    console = Console()
    table = Table(title='Recce Preset Checks', box=box.HORIZONTALS, title_style='bold dark_orange3')
    table.add_column('Name')
    table.add_column('Type')
    table.add_column('Description')
    for check in checks:
        is_check = False
        name = check.get('name')
        description = check.get('description', '')
        check_type = check.get('type')
        check_params = check.get('params', {})
        check_options = check.get('view_options', {})

        create_check_without_run(name, description, check_type, check_params, check_options, is_preset=True,
                                 is_checked=is_check)
        table.add_row(name, check_type.replace('_', ' ').title(), description.strip())
    console.print(table)


def schema_diff_should_be_approved(check_params: dict) -> bool:
    try:
        context = default_context()

        if 'node_id' in check_params:
            # If the node_id is provided, then use it
            if isinstance(check_params['node_id'], str):
                selected_node_ids = [check_params['node_id']]
            else:
                selected_node_ids = check_params.get('node_id', [])
        else:
            # Otherwise, select the nodes based on the select/exclude/packages/view_mode
            selected_node_ids = context.adapter.select_nodes(
                select=check_params.get('select'),
                exclude=check_params.get('exclude'),
                packages=check_params.get('packages'),
                view_mode=check_params.get('view_mode'),
            )

        selected_node_ids = [node for node in selected_node_ids if not node.startswith('test.')]

        def _get_selected_node_columns_from_lineage(lineage, node_ids: list[str]):
            nodes = {}
            for node_id, node in lineage.get('nodes', {}).items():
                if node_id in node_ids:
                    nodes[node_id] = node.get('columns', {})
            return nodes

        base_nodes = _get_selected_node_columns_from_lineage(context.get_lineage(base=True), selected_node_ids)
        curr_nodes = _get_selected_node_columns_from_lineage(context.get_lineage(base=False), selected_node_ids)
        diff = DeepDiff(base_nodes, curr_nodes, ignore_order=True)

        # If the diff is empty, then the check should be approved
        if bool(diff) is False:
            return True
    except Exception:
        # If there is any error, then the check should not be approved
        pass

    return False


def run_should_be_approved(run):
    if run.type == RunType.ROW_COUNT_DIFF:
        # If the run has an error, then the check should not be approved
        if run.error is not None:
            return False
        # If the row count are exactly the same, then the check should be approved
        for column, row_count_result in run.result.items():
            if row_count_result['base'] != row_count_result['curr']:
                return False
        return True
    return False


async def execute_preset_checks(preset_checks: list) -> (int, List[dict]):
    """
    Execute the preset checks
    """
    console = Console()
    rc = 0
    failed_checks = []
    table = Table(title='Recce Preset Checks', box=box.HORIZONTALS, title_style='bold dark_orange3')
    table.add_column('Status')
    table.add_column('Name')
    table.add_column('Type')
    table.add_column('Execution Time')
    table.add_column('Failed Reason')

    # Purge the existing preset checks before running the new ones
    purge_preset_checks()

    # Execute the preset checks
    for check in preset_checks:
        run = None
        check_name = check.get('name')
        check_type = check.get('type')
        check_description = check.get('description', '')
        check_params = check.get('params') if check.get('params') else {}
        check_options = check.get('view_options', {})

        try:
            # verify the check
            if check_type not in [e.value for e in RunType]:
                raise ValueError(f"Invalid check type: {check_type}")

            start = time.time()
            if check_type in ['schema_diff']:
                is_check = schema_diff_should_be_approved(check_params)
                create_check_without_run(check_name, check_description, check_type, check_params, check_options,
                                         is_preset=True, is_checked=is_check)
            else:
                run, future = submit_run(check_type, params=check_params)
                await future
                is_check = run_should_be_approved(run)
                create_check_from_run(run.run_id, check_name, check_description, check_options, is_preset=True,
                                      is_checked=is_check)

            end = time.time()
            table.add_row('[[green]Success[/green]]', check_name, check_type.replace('_', ' ').title(),
                          f'{end - start:.2f} seconds', 'N/A')
        except Exception as e:
            rc = 1
            if run is None:
                table.add_row('[[red]Error[/red]]', check_name, check_type.replace('_', ' ').title(), 'N/A', str(e))
                failed_checks.append({
                    'check_name': check_name,
                    'check_type': check_type,
                    'check_description': check_description,
                    'failed_type': 'error',
                    'failed_reason': str(e)
                })
            else:
                create_check_from_run(run.run_id, check_name, check_description, check_options, is_preset=True)
                table.add_row('[[red]Failed[/red]]', check_name, check_type.replace('_', ' ').title(), 'N/A', run.error)
                failed_checks.append({
                    'check_name': check_name,
                    'check_type': check_type,
                    'check_description': 'N/A',
                    'failed_type': 'failed',
                    'failed_reason': run.error
                })

    console.print(table)
    return rc, failed_checks


async def execute_state_checks(checks: list) -> (int, List[dict]):
    """
    Execute the checks from loaded state
    """
    console = Console()
    rc = 0
    failed_checks = []
    table = Table(title='Recce Checks', box=box.HORIZONTALS, title_style='bold dark_orange3')
    table.add_column('Status')
    table.add_column('Name')
    table.add_column('Type')
    table.add_column('Execution Time')
    table.add_column('Failed Reason')

    # Execute loaded checks
    for check in checks:
        run = None
        check_id = check.check_id
        check_name = check.name
        check_type = check.type.value
        check_description = check.description
        check_params = check.params if check.params else {}
        if check.is_checked:
            check.is_checked = False
            check.updated_at = datetime.now(tz=timezone.utc).replace(microsecond=0)

        try:
            # verify the check
            if check_type not in [e.value for e in RunType]:
                raise ValueError(f"Invalid check type: {check_type}")

            start = time.time()
            if check_type not in ['schema_diff']:
                run, future = submit_run(check_type, params=check_params, check_id=check_id)
                await future

            end = time.time()
            table.add_row('[[green]Success[/green]]', check_name, check_type.replace('_', ' ').title(),
                          f'{end - start:.2f} seconds', 'N/A')
        except Exception as e:
            rc = 1
            if run is None:
                table.add_row('[[red]Error[/red]]', check_name, check_type.replace('_', ' ').title(), 'N/A', str(e))
                failed_checks.append({
                    'check_name': check_name,
                    'check_type': check_type,
                    'check_description': check_description,
                    'failed_type': 'error',
                    'failed_reason': str(e)
                })
            else:
                table.add_row('[[red]Failed[/red]]', check_name, check_type.replace('_', ' ').title(), 'N/A', run.error)
                failed_checks.append({
                    'check_name': check_name,
                    'check_type': check_type,
                    'check_description': 'N/A',
                    'failed_type': 'failed',
                    'failed_reason': run.error
                })

    console.print(table)
    return rc, failed_checks


def process_failed_checks(failed_checks: List[dict], error_log=None):
    from py_markdown_table.markdown_table import markdown_table
    failed_check_table = []
    for check in failed_checks:
        name = check.get('check_name')
        check_type = check.get('check_type')
        failed_type = check.get('failed_type')
        failed_reason = check.get('failed_reason')
        failed_check_table.append({
            'Name': name,
            'Type': check_type,
            'Kind of Failed': failed_type,
            'Failed Reason': failed_reason.replace('\n', ' ')
        })

    content = '# Recce Runc Failed Checks\n'
    content += markdown_table(failed_check_table).set_params(quote=False, row_sep='markdown').get_markdown()

    if error_log:
        with open(error_log, 'w') as f:
            f.write(content)
        print(f"The failed checks are stored at '{error_log}'")
    else:
        print(content, file=sys.stderr)


async def cli_run(output_state_file: str, **kwargs):
    """The main function of 'recce run' command. It will execute the default runs and store the state."""
    console = Console()
    error_log = kwargs.get('error_log')
    if kwargs.get('sqlmesh', False):
        console.print("[[red]Error[/red]] SQLMesh adapter is not supported.")
        sys.exit(1)

    from recce.core import load_context
    ctx = load_context(**kwargs)

    is_skip_query = kwargs.get('skip_query', False)

    # Prepare the artifact by collecting the lineage
    console.rule("DBT Artifacts")
    from recce.adapter.dbt_adapter import DbtAdapter
    dbt_adaptor: DbtAdapter = ctx.adapter
    dbt_adaptor.print_lineage_info()

    # Execute the preset checks
    rc = 0
    if ctx.state_loader.state is None:
        preset_checks = RecceConfig().get('checks')
        if is_skip_query or preset_checks is None or len(preset_checks) == 0:
            # Skip the preset checks
            pass
        else:
            console.rule("Preset checks")
            _, failed_checks = await execute_preset_checks(preset_checks)
            if failed_checks:
                console.print("[[yellow]Warning[/yellow]] Preset checks failed. Please see the failed reason.")
                process_failed_checks(failed_checks, error_log)
    else:
        state_checks = ctx.state_loader.state.checks
        if is_skip_query or state_checks is None or len(state_checks) == 0:
            # Skip the checks in the state
            pass
        else:
            console.rule("Checks")
            _, failed_checks = await execute_state_checks(state_checks)
            if failed_checks:
                console.print("[[yellow]Warning[/yellow]] Checks failed. Please see the failed reason.")
                process_failed_checks(failed_checks, error_log)

    from recce.event import log_load_state
    log_load_state(command='run')

    # Export the state
    console.rule("Export state")
    ctx.state_loader.state_file = output_state_file
    msg = ctx.state_loader.export(ctx.export_state())
    console.print(msg)

    summary_path = kwargs.get('summary')
    if summary_path:
        dirs = os.path.dirname(summary_path)
        if dirs:
            os.makedirs(dirs, exist_ok=True)
        with open(summary_path, 'w') as f:
            f.write(generate_markdown_summary(ctx))
        console.print(f"The summary is stored at '{summary_path}'")

    return rc
